# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Test Order-based procflow WorkflowStepDefinition Model."""


# Python Standard Libraries
import copy

# Third-Party Libraries
from pydantic import ValidationError
import pytest

# GeoIPS Libraries
from geoips.pydantic import workflows


# def test_bad_workflow_step_definition_model_
#
def test_bad_workflow_step_definition_model_empty_value_kind():
    """Test WorkflowStepDefinitionModel with empty 'kind'."""
    with pytest.raises(ValidationError) as exec_info:
        workflows.WorkflowStepDefinitionModel(
            kind="", name="abi_netcdf", arguments={"variables": ["B14BT"]}
        )

    error_info = exec_info.value.errors()
    assert any(
        "Invalid input: 'kind' cannot be empty." in err["msg"] for err in error_info
    )


# This test is retained as it would be used for any PluginArgumentsModel
# where one or more attributes are required.
# ReaderArgumentsModel has all fields as optional or some default value

# def test_bad_workflow_step_definition_model_empty_arguments(valid_step_data):
#     """Tests _validate_plugin_arguments() model validator."""
#     invalid_step_data = copy.deepcopy(valid_step_data)
#     invalid_step_data["arguments"] = {}
#     print("invalid step data \t", invalid_step_data)
#     with pytest.raises(ValidationError) as exec_info:
#         workflows.WorkflowStepDefinitionModel(**invalid_step_data)

#     error_info = exec_info.value.errors()
#     assert any("argument model" in err["msg"] for err in error_info)


def test_good_workflow_step_definition_model_valid_step(valid_step_data):
    """Tests WorkflowStepDefinitionModel with valid data."""
    # creating an instance of PSDModel
    model = workflows.WorkflowStepDefinitionModel(**valid_step_data)

    assert model.kind == "reader"
    assert model.name == "abi_netcdf"
    assert model.arguments == {
        "area_def": "None",
        "variables": ["None"],
        "metadata_only": False,
        "self_register": ["None"],
    }


def test_bad_workflow_step_definition_model_validator_empty_input():
    """Tests WorkflowStepDefinitionModel with empty input."""
    with pytest.raises(ValidationError) as exec_info:
        workflows.WorkflowStepDefinitionModel()

    error_info = exec_info.value.errors()

    # Pydantic reports both kind and name as missing fields since they are
    # required. len(err_info) is set to 2 for this reason.
    assert len(error_info) == 2
    # , after element makes it a tuple otherwise it's a string
    expected_missing_fields = {("kind",), ("name",)}
    actual_missing_fields = {tuple(err["loc"]) for err in error_info}
    assert expected_missing_fields.issubset(actual_missing_fields)


def test_bad_workflow_step_definition_model_validator_invalid_plugin_name(
    valid_step_data,
):
    """Tests WorkflowStepDefinitionModel validator against invalid plugin name."""
    invalid_data = copy.deepcopy(valid_step_data)
    invalid_data["name"] = ""
    with pytest.raises(ValidationError) as exec_info:
        workflows.WorkflowStepDefinitionModel(**invalid_data)

    error_info = exec_info.value.errors()

    assert error_info[0]["loc"] == ()
    assert error_info[0]["msg"].startswith("Value error, ")
    assert "Must be one of " in error_info[0]["msg"]


def test_bad_workflow_step_definition_model_validate_kind(
    valid_step_data,
):
    """Tests WorkflowStepDefinitionModel validator against invalid plugin kind."""
    invalid_data = copy.deepcopy(valid_step_data)
    invalid_data["kind"] = "invalid_plugin_kind"

    with pytest.raises(ValidationError) as exec_info:
        workflows.WorkflowStepDefinitionModel(**invalid_data)

    error_info = exec_info.value.errors()
    assert "Invalid plugin kind" in error_info[0]["msg"]


def test_bad_workflow_step_definition_model_additional_field(valid_step_data):
    """Tests ReaderArgumentsModel with additional field."""
    invalid_data = copy.deepcopy(valid_step_data)
    invalid_data["unexpected_field"] = "unexpected_value"

    with pytest.raises(ValidationError) as exec_info:
        workflows.WorkflowStepDefinitionModel(**invalid_data)

    error_info = exec_info.value.errors()
    assert any(
        err["type"] == "extra_forbidden" for err in error_info
    ), "expected 'extra_forbidden' error type"
    assert "unexpected_field" in str(
        exec_info.value
    ), "Unexpected field should be rejected by the WorkflowStepDefinitionModel"


def test_bad_workflow_step_definition_model_invalid_field_type(
    mocker, valid_reader_arguments_model_data, valid_step_data
):
    """Tests WorkflowStepDefinitionModel for invalid field type instantiation."""

    def mock_init(self, **kwargs):
        self.__dict__.update(kwargs)

    mocker.patch.object(workflows.WorkflowStepDefinitionModel, "__init__", mock_init)

    invalid_test_data_PSDModel = {
        "type": 123,
        "name": 123,
        "arguments": valid_reader_arguments_model_data,
    }
    model = workflows.WorkflowStepDefinitionModel(**invalid_test_data_PSDModel)

    assert model.type == 123, "Expected 'type' to retain the invalid value"
    assert model.name == 123, "Expected 'name' to retain the invalid value"



"""Unit test module for the CLI command 'geoips config create-registries'."""

import pytest

from geoips.commandline.commandline_interface import main
from geoips.errors import PluginRegistryError
from tests.unit_tests.commandline.cli_top_level_tester import (
    generate_id,
    check_valid_command_using_monkeypatch,
)

valid_args = [
    ["geoips", "config", "create-registries"],
    ["geoips", "config", "create-registries", "-s", "yaml"],
    ["geoips", "config", "create-registries", "-p", "geoips"],
    ["geoips", "config", "create-registries", "-n", "geoips.plugin_packages"],
    [
        "geoips",
        "config",
        "create-registries",
        "-n",
        "geoips.plugin_packages",
        "-p",
        "geoips",
    ],
]
valid_expected = [
    {
        "log_level": "interactive",
        "command": "geoips config create-registries",
        "namespace": "geoips.plugin_packages",
        "packages": None,
        "save_type": "json",
    },
    {
        "log_level": "interactive",
        "command": "geoips config create-registries",
        "namespace": "geoips.plugin_packages",
        "packages": None,
        "save_type": "yaml",
    },
    {
        "log_level": "interactive",
        "command": "geoips config create-registries",
        "namespace": "geoips.plugin_packages",
        "packages": ["geoips"],
        "save_type": "json",
    },
    {
        "log_level": "interactive",
        "command": "geoips config create-registries",
        "namespace": "geoips.plugin_packages",
        "packages": None,
        "save_type": "json",
    },
    {
        "log_level": "interactive",
        "command": "geoips config create-registries",
        "namespace": "geoips.plugin_packages",
        "packages": ["geoips"],
        "save_type": "json",
    },
]
invalid_args = [
    ["geoips", "config", "create-registries", "-p", "fake_package"],
    ["geoips", "config", "create-registries", "-n", "fake.plugin_packages"],
    ["geoips", "config", "create-registries", "-s", "invalid_save_type"],
]


@pytest.mark.parametrize("cli_args", [args for args in invalid_args], ids=generate_id)
def test_create_registries_invalid(monkeypatch, cli_args):
    """Test the CLI command 'geoips config create-registries' using invalid args.

    Parameters
    ----------
    cli_args: list[str]
        - A list of strings representing the CLI command being ran.
    """
    monkeypatch.setattr("sys.argv", cli_args)
    if "-n" in cli_args or "-p" in cli_args:
        error = PluginRegistryError
    else:
        # This occurs when 'save_type' is not one of 'json' or 'yaml'. Argparse
        # eventually calls sys.exit(2, msg), which is caught here.
        error = SystemExit
    with pytest.raises(error):
        main()


@pytest.mark.parametrize(
    "cli_args,expected",
    [(args, expected) for args, expected in zip(valid_args, valid_expected)],
    ids=generate_id,
)
def test_create_registries_valid(monkeypatch, cli_args, expected):
    """Test the CLI command 'geoips config create-registries' using valid args.

    Parameters
    ----------
    cli_args: list[str]
        - A list of strings representing the CLI command being ran.
    expected: dict
        - A dictionary of expected argument values that are returned from the command
          being ran from args.
    """
    check_valid_command_using_monkeypatch(monkeypatch, cli_args, expected)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Projection information for setting up pyresample area definitions."""

projections_avail = {
    ("aeqd", "aeqd", "Azimuthal Equidistant"): 0,
    ("poly", "poly", "Polyconic"): 0,
    ("gnom", "gnom", "Gnomonic"): 0,
    ("moll", "moll", "Mollweide"): 1,
    ("tmerc", "tmerc", "Transverse Mercator"): 0,
    ("nplaea", "laea", "North-Polar Lambert Azimuthal"): 4,
    ("gall", "gall", "Gall Stereographic Cylindrical"): 0,
    ("mill", "mill", "Miller Cylindrical"): 0,
    ("merc", "merc", "Mercator"): 0,
    ("stere", "stere", "Stereographic"): 0,
    ("npstere", "stere", "North-Polar Stereographic"): 4,
    ("hammer", "hammer", "Hammer"): 1,
    ("geos", "geos", "Geostationary"): 2,
    ("nsper", "nsper", "Near-Sided Perspective"): 2,
    ("vandg", "vandg", "van der Grinten"): 0,
    ("laea", "laea", "Lambert Azimuthal Equal Area"): 0,
    ("mbtfpq", "mbtfpq", "McBryde-Thomas Flat-Polar Quartic"): 1,
    ("sinu", "sinu", "Sinusoidal"): 1,
    ("spstere", "stere", "South-Polar Stereographic"): 4,
    # lcc does not geolocate correctly
    ("lcc", "lcc", "Lambert Conformal"): 0,
    ("npaeqd", "npaeqd", "North-Polar Azimuthal Equidistant"): 4,
    ("ease_sh", "ease_sh", "Antarctic EASE Grid"): 4,
    ("ease_nh", "ease_nh", "Arctic EASE grid"): 4,
    ("eqdc", "eqdc", "Equidistant Conic"): 0,
    ("cyl", "eqc", "Equidistant Cylindrical"): 2,
    ("eqc", "eqc", "Equidistant Cylindrical"): 2,
    ("omerc", "omerc", "Oblique Mercator"): 0,
    # aea does not geolocate correctly
    ("aea", "aea", "Albers Equal Area"): 0,
    ("spaeqd", "spaeqd", "South-Polar Azimuthal Equidistant"): 4,
    ("ortho", "ortho", "Orthographic"): 3,
    ("cass", "cass", "Cassini-Soldner"): 0,
    ("splaea", "laea", "South-Polar Lambert Azimuthal"): 4,
    ("robin", "robin", "Robinson"): 1,
}


def get_projection(name):
    """Get a dictionary of projection names containing the specified keys.

    Dictionary keys:
        * name:       the basemap projection short name
        * p4name:     the Proj4 projection name
        * longname:   a long name describing the projection
        * type:       an integer indicating how the projection must be set up

    The type field tells the program which arguments will be useful to a given
    projection

    * Can use corner lats and lons or center lats and lons with width and height
    * Ignores corner lats and lons and width/height arguments.  Uses center lat/lon
    * Can use corner lats and lons or corner coordinates in the local projection
      space, but ignores all other location parameters.
    """
    proj_info = None
    for key, val in projections_avail.items():
        try:
            lowercase_name = name.text.lower()
        except AttributeError:
            lowercase_name = name.lower()
        if lowercase_name == key[0].lower():
            proj_info = {
                "name": key[0],
                "p4name": key[1],
                "longname": key[2],
                "type": val,
            }
            break
    if proj_info is None:
        raise TypeError(f"{name} projection is not defined in geoips.")
    if proj_info["p4name"] is None:
        raise TypeError(f"{name} projection is not defined in Proj4.")
    return proj_info



"""Testing module for Pydantic SectorPluginModel."""

from copy import deepcopy
from importlib.resources import files

import pytest
import yaml

from geoips.pydantic.sectors import SectorPluginModel
from tests.unit_tests.pydantic.utils import (
    PathDict,
    load_test_cases,
    validate_bad_plugin,
    validate_good_plugin,
)

interface = "sectors"


test_cases = load_test_cases(interface)
good_yaml = yaml.safe_load(
    open(str(files("geoips") / "plugins/yaml/sectors/static/korea.yaml"), mode="r")
)
good_yaml["abspath"] = str(files("geoips") / "plugins/yaml/sectors/static/korea.yaml")
good_yaml["relpath"] = "plugins/yaml/sectors/static/korea.yaml"
good_yaml["package"] = "geoips"


@pytest.fixture
def good_sector():
    """Return a consistent dictionary that is a valid GeoIPS sector plugin."""
    # Make the loading code only occur once, return a copy every time
    return PathDict(deepcopy(good_yaml))


@pytest.mark.parametrize(
    "good_sector",
    [pytest.param("good_sector", id="good_sector")],
    indirect=True,
)
def test_good_sector(good_sector):
    """Assert that a well formatted sector plugin is valid.

    Parameters
    ----------
    good_sector: dict
        - A dictionary representing a valid sector plugin.
    """
    validate_good_plugin(good_sector, SectorPluginModel)


@pytest.mark.parametrize("test_tup", test_cases.values(), ids=list(test_cases.keys()))
def test_bad_sector_plugins(good_sector, test_tup):
    """Perform validation on static sector plugins, including failing cases.

    Parameters
    ----------
    good_sector: dict
        - A dictionary representing a sector plugin that is valid.
    test_tup:
        - A tuple formatted (key, value, class, err_str), formatted (str, any, str, str)
          used to run and validate tests.
    """
    validate_bad_plugin(good_sector, test_tup, SectorPluginModel)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Filename formatter for TC-specific text windspeed outputs."""

import logging

from os.path import join as pathjoin

from geoips.filenames.base_paths import PATHS as gpaths

interface = "filename_formatters"
family = "xarray_metadata_to_filename"
name = "text_winds_tc_fname"

LOG = logging.getLogger(__name__)


def call(xarray_obj, extension=".txt", basedir=gpaths["TCWWW"], output_dict=None):
    """Create TC-specific text windspeed filename.

    See Also
    --------
    geoips.plugins.modules.filename_formatters.text_winds_tc_fname.
        assemble_windspeeds_text_tc_fname
        Shared utility to facilitate creating multiple similar filenames
        from the same code.
    """
    area_def = xarray_obj.area_definition
    return assemble_windspeeds_text_tc_fname(
        basedir=basedir,
        tc_area_id=area_def.area_id,
        tc_year=int(area_def.sector_info["storm_year"]),
        tc_basin=area_def.sector_info["storm_basin"],
        tc_stormnum=int(area_def.sector_info["storm_num"]),
        source_name=xarray_obj.source_name,
        platform_name=xarray_obj.platform_name,
        product_datetime=xarray_obj.start_datetime,
        data_provider=xarray_obj.data_provider,
        extension=extension,
        output_dict=output_dict,
        sector_info=area_def.sector_info,
    )


def assemble_windspeeds_text_tc_fname(
    basedir,
    tc_area_id,
    tc_year,
    tc_basin,
    tc_stormnum,
    source_name,
    platform_name,
    product_datetime,
    data_provider,
    extension=".txt",
    output_dict=None,
    sector_info=None,
):
    """Produce full output product path from product / sensor specifications.

    Parameters
    ----------
    basedir : str
         base directory
    tc_year : int
         Full 4 digit storm year
    tc_basin : str
         2 character basin designation
            SH Southern Hemisphere
            WP West Pacific
            EP East Pacific
            CP Central Pacific
            IO Indian Ocean
            AL Atlantic
    tc_stormnum : int
        2 digit storm number
            90 through 99 for invests
            01 through 69 for named storms
    platform_name : str
        Name of platform (satellite)
    product_datetime : datetime
        Start time of data used to generate product

    Returns
    -------
    str:
        full path of output filename of the format:
          <basedir>/tc<tc_year>/<tc_basin>/<tc_basin><tc_stormnum><tc_year>/
            txt/<source_name>_<platform_name>_surface_winds_<data_provider>_
            <YYYYMMDDHHMN>

    Examples
    --------
    >>> startdt = datetime.strptime('20200216T001412', '%Y%m%dT%H%M%S')
    >>> assemble_windspeeds_text_tc_fname('/outdir',
    ...     2020,
    ...     'SH',
    ...     16,
    ...     'smap-spd',
    ...     'smap',
    ...     startdt,
    ...     'remss')
    '/outdir/tc2020/SH/SH162020/txt/
    """
    from geoips.plugins.modules.filename_formatters.utils.tc_file_naming import (
        tc_storm_basedir,
    )

    path = pathjoin(
        tc_storm_basedir(
            basedir,
            tc_year,
            tc_basin,
            tc_stormnum,
            output_dict=output_dict,
            sector_info=sector_info,
        ),
        "txt",
    )
    fname = "_".join(
        [
            source_name,
            "surface_winds",
            data_provider,
            platform_name,
            tc_area_id,
            product_datetime.strftime("%Y%m%d%H%M"),
        ]
    )
    if extension is not None:
        fname = f"{fname}{extension}"
    return pathjoin(path, fname)



"""Generic testing utilities used for pydantic unit testing."""

# from collections import UserDict
import logging
from importlib.resources import files
import os
import pytest

import numpy as np
from pydantic import ValidationError
import yaml

from geoips import interfaces
from geoips import pydantic as gpydan


LOG = logging.getLogger(__name__)


class PathDict(dict):
    """Custom dictionary class used to access GeoIPS plugins using paths.

    Where a 'path' is a key which can include slashes ('/') for easy access to the
    values contained in nested dictionaries.

    I.e. some_dict["key1/key2/key3"] = some_val

    This is essentially the same thing as a dict object besides the __setitem__ method.
    Not inheriting from that class to avoid unexpected consequences.
    """

    def __setitem__(self, key, value):
        """Set an item under key with value 'value'.

        Custom implementation of dict's __setitem__, which makes use of paths to
        set an item which may be multiple levels deep.

        Parameters
        ----------
        key: str
            - The path to the attribute to set
        value: any
            - The value to set the attribute to.
        """
        keys = key.split("/")
        if len(keys) > 1:
            current = self
            for subkey in keys[:-1]:
                current = current.setdefault(subkey, {})
            dict.__setitem__(current, keys[-1], value)
        else:
            dict.__setitem__(self, key, value)


def load_test_cases(interface_name):
    """Load a set of test cases used to validate pydantic model(s).

    This can either be a top level model, such as SectorPluginModel, or a component
    of that model.

    Parameters
    ----------
    interface_name: str
        - The name of the interface that's going to be tested.

    Returns
    -------
    test_cases: dict
        - The dictionary of test cases used to validate your model.
    """
    fpath = f"{os.path.dirname(__file__)}/{interface_name}/test_cases.yaml"
    if not os.path.exists(fpath):
        raise FileNotFoundError(
            f"Error: No test cases file could be found. Expected {fpath} but it did not"
            " exist. Please create this file and rerun your tests."
        )
    with open(fpath, "r") as fo:
        test_cases = yaml.safe_load(fo)

    for id, val in test_cases.items():
        for key in list(val.keys()):
            if key not in ("key", "val", "cls", "err_str", "warn_match"):
                error = (
                    f"ERROR: test_case '{id}' has item with key '{key}' which is an "
                    "invalid test case key. We only support the full set of keys "
                    "[key, val, cls, err_str] at this moment."
                )
                raise RuntimeError(error)
    return test_cases


def load_geoips_yaml_plugin(interface_name, plugin_name):
    """Load a GeoIPS YAML plugin via yaml.safe_load, not interface.get_plugin.

    This will be used until we convert all of our schema to pydantic. If we load a
    plugin via <interface>.get_plugin, this sometimes will cause errors due to different
    validation protocols.

    Parameters
    ----------
    interface_name: str
        - The name of the GeoIPS plugin's interface.
    plugin_name: str
        - The name of the plugin of type 'interface_name'.

    Returns
    -------
    yam: dict
        - A dictionary object representing the yaml plugin requested.
    """
    yam_ints = interfaces.list_available_interfaces()["yaml_based"]
    if interface_name not in yam_ints:
        raise KeyError(
            f"'{interface_name}' is not a valid yaml interface. Please select an "
            f"interface out of any of the following {yam_ints}."
        )
    interface = getattr(interfaces, interface_name)
    registry = interface.plugin_registry.registered_plugins["yaml_based"][
        interface_name
    ]
    if isinstance(plugin_name, tuple):
        # Likely a product. Do a nested search for the plugin.
        entry = registry[plugin_name[0]][plugin_name[1]]
    else:
        entry = registry[plugin_name]

    relpath = entry["relpath"]
    abspath = str(files("geoips") / relpath)
    package = "geoips"

    yam = yaml.safe_load(open(abspath, "r"))
    # Append these attributes to the plugin for further validation. 'get_plugin'
    # already does this, but since we're loading in a different manner, we do this
    # manually for the time being.
    yam["abspath"] = abspath
    yam["relpath"] = relpath
    yam["package"] = package

    return yam


def validate_good_plugin(good_plugin, plugin_model):
    """Assert that a well formatted plugin is valid.

    Parameters
    ----------
    good_plugin: dict
        - A dictionary representing a valid plugin.
    plugin_model: instance or child of geoips.pydantic.bases.PluginModel
        - The pydantic-based model used to validate this plugin.
    """
    plugin_model(**good_plugin)


def validate_bad_plugin(good_plugin, test_tup, plugin_model):
    """Perform validation on any GeoIPS plugin, ensuring correct ValidationErrors occur.

    Current supported plugins include readers and sectors. More to come in the future,
    and this function might change because of that. This is a beta method for the time
    being.

    Parameters
    ----------
    good_plugin: dict
        - A dictionary representing a plugin that is valid.
    test_tup:
        - A tuple formatted (key, value, class, err_str), formatted (str, any, str, str)
          used to run and validate tests.
    plugin_model: instance or child of geoips.pydantic.bases.PluginModel
        - The pydantic-based model used to validate this plugin.
    """
    bad_plugin = good_plugin
    key = test_tup["key"]
    val = test_tup["val"]
    failing_model = test_tup["cls"]

    err_str = test_tup.get("err_str")
    warn_match = test_tup.get("warn_match")

    if err_str is not None and warn_match is not None:
        raise ValueError("Only one of 'err_str' or 'warn_match' should be set.")
    elif err_str is None and warn_match is None:
        raise ValueError("At least one of 'err_str' or 'warn_match' must be provided.")

    # ValidationErrors won't occur for these fields at the moment. Assert that the
    # plugin validates for the time being
    if key in ["abspath", "relpath", "package"]:
        bad_plugin.pop(key)
        plugin_model(**bad_plugin)
    # Otherwise, set the bad value in the plugin, and test for ValidationErrors
    else:
        bad_plugin[key] = val

        if warn_match:
            with pytest.warns(FutureWarning, match=warn_match):
                plugin_model(**bad_plugin)
        else:
            try:
                plugin_model(**bad_plugin)
            except ValidationError as e:
                # The code below assumes that your test only raised one error. That's
                # how we've structured testing for the time being. In the case that one
                #  or more errors are reported, we default to the last error of the
                # failing model reported, or, if no failing model could be associated
                # with this error, we just default to the last error reported.
                errors = e.errors()
                if len(e.errors()) > 1:
                    val_err = attempt_to_associate_model_with_error(
                        failing_model, errors
                    )
                else:
                    val_err = errors[0]
                # In testing, it seems that the last 'loc' is always the failing
                # attribute
                bad_field = val_err["loc"][-1]
                err_msg = val_err["msg"]
                # Find the module which contains the failing model. I.e. PluginModel in
                # geoips.pydantic.bases
                module = None
                for mod in gpydan._modules:
                    if failing_model in gpydan._classes[mod]:
                        module = mod
                        break
                    elif hasattr(gpydan._modules[mod], failing_model):
                        # This behavior occurs for the 'ColorType' attribute, which is
                        # a type instance but not actually a pydantic class. Skip the
                        # field assertion below
                        module = "pass"
                        break
                if module != "pass":
                    # Assert that the failing field was found in the model expected to
                    # fail
                    assert (
                        module
                        and bad_field
                        in getattr(gpydan._modules[module], failing_model).model_fields
                    )
                if err_str:
                    # Assert that the error string provided is in the error message
                    # returned or equal to the error message returned.
                    assert err_str in err_msg or err_str == err_msg


def attempt_to_associate_model_with_error(failing_model, errors):
    """Attempt to associate the correct error with the model that should be failing.

    This occurs when more than one error was raised for a given test. Assuming that
    two fields were not incorrect, only one, then this means that a field is one of a
    Union of models and/or other fields. Attempt to locate the correct error based on
    the model provided.

    Parameters
    ----------
    failing_model: str
        - The name (case sensitive) of the model that we expect to raise this error
    errors: pydantic.ValidationError.errors() -- list
        - A list of errors containing information on why a field failed to validate.
    """
    val_err_inds = []
    for err in errors:
        failing_model_found = False
        for loc in err["loc"]:
            if loc == failing_model:
                failing_model_found = True
                break
        val_err_inds.append(failing_model_found)
    val_err_inds = np.array(val_err_inds)
    # If no errors could be associated with the failing model, just default
    # to the last error reported
    if not np.any(val_err_inds):
        val_err = np.array(errors)[-1]
    # Otherwise, choose the last error associated with the failing model.
    else:
        val_err = np.array(errors)[val_err_inds][-1]

    return val_err



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Read SEVIRI hrit data.

Notes
-----
   1) At present, this reader does not work for High Resolution Visible data,
      which is ignored. Additionally, to ease generation of geolocation fields,
      datasets are assumed to be square and centered at their sub longitude.

20170330 MLS Try to only decompress what we need (VERY filename dependent),
  make scifile and hrit channel names match (more filename dependence),
  don't try to decompress/open file for import_metadata (more filename
  dependence for time). satpy requires time to open file, and requires standard
  (decompressed) filenames, so built in filename dependence by using satpy.
"""
# Python Standard Libraries
import os
import logging
import numpy as np
from geoips.plugins.modules.readers.utils.hrit_reader import HritFile, HritError

# GeoIPS Libraries
from geoips.errors import NoValidFilesError
from geoips.interfaces import readers
from geoips.filenames.base_paths import PATHS as gpaths
from geoips.utils.context_managers import import_optional_dependencies
from geoips.plugins.modules.readers.utils.geostationary_geolocation import (
    get_geolocation,
)

LOG = logging.getLogger(__name__)

with import_optional_dependencies(loglevel="info"):
    """Attempt to import a package and print to LOG.info if the import fails."""
    import numexpr as ne

try:
    NPROC = 6
    ne.set_num_threads(NPROC)
except NameError:
    err_str = "Failed ne.set_num_threads in seviri_hrit.py. "
    err_str += "If you need numexpr, install it."
    LOG.info(err_str)


# These should be added to the data file object
BADVALS = {
    "Off_Of_Disk": -999.9,
}

VIS_CALIB = {
    "msg1": {"B01": 65.2296, "B02": 73.0127, "B03": 62.3715, "B12": 78.7599},
    "msg2": {"B01": 65.2065, "B02": 73.1869, "B03": 61.9923, "B12": 79.0113},
    "msg3": {"B01": 65.5148, "B02": 73.1807, "B03": 62.0208, "B12": 78.9416},
    "msg4": {"B01": 65.2656, "B02": 73.1692, "B03": 61.9416, "B12": 79.0035},
}

IR_CALIB = {
    "msg1": {
        "B04": {"wn": 2567.330, "a": 0.9956, "b": 3.410},  # IR3.9 Water vapour channel
        "B05": {"wn": 1598.103, "a": 0.9962, "b": 2.218},  # IR6.2 Water vapour channel
        "B06": {"wn": 1362.081, "a": 0.9991, "b": 0.478},  # IR7.3 Water vapour channel
        "B07": {"wn": 1149.069, "a": 0.9996, "b": 0.179},  # IR8.7 Atmospheric window
        "B08": {"wn": 1034.343, "a": 0.9999, "b": 0.060},  # IR9.7 Ozone channel
        "B09": {"wn": 930.647, "a": 0.9983, "b": 0.625},  # IR10.8 Atmospheric window
        "B10": {"wn": 839.660, "a": 0.9988, "b": 0.397},  # IR12.0 Atmospheric window
        "B11": {"wn": 752.387, "a": 0.9981, "b": 0.578},
    },  # IR13.4 Carbon dioxide channel
    "msg2": {
        "B04": {"wn": 2568.832, "a": 0.9954, "b": 3.438},  # IR3.9 Water vapour channel
        "B05": {"wn": 1600.548, "a": 0.9963, "b": 2.185},  # IR6.2 Water vapour channel
        "B06": {"wn": 1360.330, "a": 0.9991, "b": 0.470},  # IR7.3 Water vapour channel
        "B07": {"wn": 1148.620, "a": 0.9996, "b": 0.179},  # IR8.7 Atmospheric window
        "B08": {"wn": 1035.289, "a": 0.9999, "b": 0.056},  # IR9.7 Ozone channel
        "B09": {"wn": 931.700, "a": 0.9983, "b": 0.640},  # IR10.8 Atmospheric window
        "B10": {"wn": 836.445, "a": 0.9988, "b": 0.408},  # IR12.0 Atmospheric window
        "B11": {"wn": 751.792, "a": 0.9981, "b": 0.561},
    },  # IR13.4 Carbon dioxide channel
    "msg3": {
        "B04": {"wn": 2547.771, "a": 0.9915, "b": 2.9002},  # IR3.9 Water vapour channel
        "B05": {"wn": 1595.621, "a": 0.9960, "b": 2.0337},  # IR6.2 Water vapour channel
        "B06": {"wn": 1360.377, "a": 0.9991, "b": 0.4340},  # IR7.3 Water vapour channel
        "B07": {"wn": 1148.130, "a": 0.9996, "b": 0.1714},  # IR8.7 Atmospheric window
        "B08": {"wn": 1034.715, "a": 0.9999, "b": 0.0527},  # IR9.7 Ozone channel
        "B09": {"wn": 929.842, "a": 0.9983, "b": 0.6084},  # IR10.8 Atmospheric window
        "B10": {"wn": 838.659, "a": 0.9988, "b": 0.3882},  # IR12.0 Atmospheric window
        "B11": {"wn": 750.653, "a": 0.9982, "b": 0.5390},
    },  # IR13.4 Carbon dioxide channel
    "msg4": {
        "B04": {"wn": 2555.280, "a": 0.9916, "b": 2.9438},  # IR3.9 Water vapour channel
        "B05": {"wn": 1596.080, "a": 0.9959, "b": 2.0780},  # IR6.2 Water vapour channel
        "B06": {"wn": 1361.748, "a": 0.9990, "b": 0.4929},  # IR7.3 Water vapour channel
        "B07": {"wn": 1148.130, "a": 0.9996, "b": 0.1731},  # IR8.7 Atmospheric window
        "B08": {"wn": 1034.851, "a": 0.9998, "b": 0.0597},  # IR9.7 Ozone channel
        "B09": {"wn": 931.122, "a": 0.9983, "b": 0.6256},  # IR10.8 Atmospheric window
        "B10": {"wn": 839.113, "a": 0.9988, "b": 0.4002},  # IR12.0 Atmospheric window
        "B11": {"wn": 748.585, "a": 0.9981, "b": 0.5635},
    },
}  # IR13.4 Carbon dioxide channel

geolocation_variable_names = [
    "latitude",
    "longitude",
    "solar_zenith_angle",
    "satellite_zenith_angle",
    "solar_azimuth_angle",
    "satellite_azimuth_angle",
]

interface = "readers"
family = "standard"
name = "seviri_hrit"


def calculate_chebyshev_polynomial(coefs, start_dt, end_dt, dt):
    """Calculate Chebyshev Polynomial."""
    start_to_end = (end_dt - start_dt).seconds
    # dt_to_end = (end_dt - dt).seconds
    start_to_dt = (dt - start_dt).seconds
    t = (start_to_dt - 0.5 * (start_to_end)) / (0.5 * (start_to_end))
    t2 = 2 * t

    # I think this is what was in the documentation
    # MSG_Level_1_5_Image_Data_Format_Description.pdf
    # p 87 earth fixed coordinate frame
    # p 124 decoding chebychev polynomial function
    d = 0
    dd = 0
    for j in range(7, 1, -1):
        save = d
        d = t2 * d - dd + coefs[j]
        dd = save
    d = t * d - dd + 0.5 * coefs[0]

    # f = -0.5 * coefs[0] # First term of Chebyshev polynomial
    # f = f + coefs[0] # second term of Chebyshev polynomial
    # f = f + coefs[1]*t # third term
    # T_k_minus_3 = 1
    # T_k_minus_2 = t
    # T_k_minus_1 = t2*T_k_minus_2 - T_k_minus_3
    # remaining terms recursively defined
    # for xcoef in coefs[3:]:
    #    f = f + xcoef * T_k_minus_1
    #    T_k_minus_3 = T_k_minus_2
    #    T_k_minus_2 = T_k_minus_1
    #    T_k_minus_1 = t2*T_k_minus_2 - T_k_minus_3

    return d


class XritError(Exception):
    """Raise exception on XritError."""

    def __init__(self, msg, code=None):
        """Initialize XritError."""
        self.code = code
        self.value = msg

    def __str__(self):
        """Return XritError string."""
        if self.code:
            return "{}: {}".format(self.code, self.value)
        else:
            return self.value


def compare_dicts(d1, d2, skip=None):
    """Compare the values in two dictionaries.

    If they are equal, return True, otherwise False
    If skip is set and contains one of the keys, skip that key
    """
    if d1.keys() != d2.keys():
        return False

    for key in d1.keys():
        if skip and key in skip:
            continue
        if d1[key] != d2[key]:
            return False
    return True


def get_top_level_metadata(fnames, sect):
    """Get top level metadata."""
    md = {}
    for fname in fnames:
        df = HritFile(fname)
        if "block_2" in df.metadata.keys():
            break
    if "GEOS" in df.metadata.get("block_2", {}).get("projection", {}):
        md["sector_name"] = "Full-Disk"
    else:
        projection = df.metadata.get("block_2", {}).get("projection", {})
        st = df.start_datetime.isoformat()
        et = df.start_datetime.isoformat()
        raise HritError(
            f"Unknown projection encountered: {projection}.\n"
            f"start_datetime={st}\n"
            f"end_datetime={et}"
        )
    md["start_datetime"] = df.start_datetime
    md["end_datetime"] = df.start_datetime
    md["data_provider"] = "nesdisstar"
    # MLS Check platform_name
    # Turn msg4_iodc into msg4.  Then pull geoips satname (meteoEU/meteoIO)
    # from utils/satellite_info.py
    msg_satname = df.annotation_metadata["platform"].lower().replace("_iodc", "")
    # Save actual satellite name (msg1 / msg4) for the coefficient tables above.
    # geoips specific platform_name should be msg-1 or msg-4
    md["satellite_name"] = msg_satname
    # from geoips.utils.satellite_info import open_satinfo
    # try:
    #     satinfo = open_satinfo(msg_satname)
    #     if hasattr(satinfo, 'geoips_satname'):
    #         msg_satname = satinfo.geoips_satname
    # except KeyError:
    #     raise HritError('Unknown satname encountered: {}'.format(msg_satname))
    md["platform_name"] = msg_satname[0:3] + "-" + msg_satname[-1]  # msg-1 or msg-4
    md["source_name"] = "seviri"
    md["source_file_names"] = [os.path.basename(fname) for fname in fnames]
    md["area_definition"] = sect
    md["sample_distance_km"] = 3.0

    return md


def get_latitude_longitude(gmd, BADVALS, area_def):
    """Generate full-disk latitudes and longitudes."""
    # Anywhere you see NOQA or noqa: F841, this is added since the variables are used
    # however they are not recognized by flake8 linter under numexpr.evaluate()
    # Constants
    pi = np.pi
    # Must include rad2deg variable, because it is used within the
    # numexpr command below.  flake8 does not recognize it as being
    # used, so must include # NOQA flag
    rad2deg = 180.0 / pi  # NOQA
    deg2rad = pi / 180.0
    Rs = 42164  # Satellite altitude (km)  # noqa: F841
    Re = 6378.1690  # Earth equatorial radius (km)
    Rp = 6356.5838  # Earth polar radius (km)
    r3 = Re**2 / Rp**2  # noqa: F841
    sd_coeff = 1737122264  # If there is a problem for MSG use 1737121856  # noqa: F841
    lon0 = gmd["lon0"]  # noqa: F841

    deg2rad = np.pi / 180.0
    x, y = np.meshgrid(
        np.arange(0, gmd["num_samples"], 1), np.arange(0, gmd["num_lines"], 1)
    )
    x = np.fliplr(x)
    y = np.fliplr(y)
    x = deg2rad * (x - gmd["sample_offset"]) / (2**-16 * gmd["sample_scale"])
    y = deg2rad * (y - gmd["line_offset"]) / (2**-16 * gmd["line_scale"])

    cos_x = np.cos(x)
    sin_x = np.sin(x)  # noqa: F841
    cos_y = np.cos(y)
    sin_y = np.sin(y)  # noqa: F841

    sd = ne.evaluate("(Rs * cos_x * cos_y)**2 - (cos_y**2 + r3 * sin_y**2) * sd_coeff")
    bad_mask = sd < 0.0
    sd[bad_mask] = 0.0
    sd **= 0.5

    # Doing inplace operations when variables are no longer needed

    # sd no longer needed
    sn = sd
    ne.evaluate("(Rs * cos_x * cos_y - sd) / (cos_y**2 + r3 * sin_y**2)", out=sn)

    # cos_x no longer needed
    s1 = cos_x
    ne.evaluate("Rs - (sn * cos_x * cos_y)", out=s1)

    # Nothing unneed
    s2 = ne.evaluate("sn * sin_x * cos_y")  # noqa: F841

    # sin_y no longer needed
    s3 = cos_y
    ne.evaluate("-sn * sin_y", out=s3)

    # sn no longer needed
    sxy = sn
    ne.evaluate("sqrt(s1**2 + s2**2)", out=sxy)

    # s3 no longer needed
    lats = s3
    ne.evaluate("rad2deg * arctan(r3 * s3 / sxy)", out=lats)

    # s1 no longer needed
    lons = s1
    ne.evaluate("rad2deg * arctan(s2 / s1) + lon0", out=lons)
    lons[lons > 180.0] -= 360

    # Set bad values
    lats[bad_mask] = BADVALS["Off_Of_Disk"]
    lons[bad_mask] = BADVALS["Off_Of_Disk"]

    return lats, lons


# def get_low_res_geolocation_args(prologue):
def _get_geolocation_metadata(prologue, file_metadata):
    """Get geolocation metadata."""
    geomet = {}
    geomet["platform_name"] = file_metadata["platform_name"]
    geomet["source_name"] = file_metadata["source_name"]
    geomet["scan_mode"] = prologue["imageDescription"]["projectionDescription"][
        "typeOfProjection"
    ]
    # Used for cached filenames
    geomet["scene"] = geomet["scan_mode"]
    geomet["lon0"] = prologue["imageDescription"]["projectionDescription"][
        "longitudeOfSSP"
    ]
    geomet["num_lines"] = prologue["imageDescription"]["referenceGridVIS_IR"][
        "numberOfLines"
    ]
    geomet["num_samples"] = prologue["imageDescription"]["referenceGridVIS_IR"][
        "numberOfColumns"
    ]
    geomet["line_scale"] = -13642337  # This will only work for low resolution data
    geomet["sample_scale"] = -13642337
    geomet["line_offset"] = geomet["num_lines"] / 2
    geomet["sample_offset"] = geomet["num_samples"] / 2
    geomet["start_datetime"] = prologue["imageAcquisition"]["plannedAcquisitionTime"][
        "trueRepeatCycleStart"
    ]
    geomet["H_m"] = 42164 * 1000.0  # Satellite altitude (m)
    geomet["roi_factor"] = 10  # roi = res_km * 1000 * roi_factor
    return geomet


def countsToRad(counts, slope, offset):
    """Convert counts to rad."""
    rad = np.full_like(counts, -999.0)
    rad[counts > 0] = offset + (slope * counts[counts > 0])
    return rad


def radToRef(rad, sun_zen, platform, band):
    """Convert Rad to Ref."""
    irrad = VIS_CALIB[platform][band]
    ref = np.full_like(rad, -999.0)
    # 0 to 1 rather than 0 to 100
    ref[rad > 0] = rad[rad > 0] / irrad
    # DO NOT REMOVE THIS STEP ALTOGETHER!!!!  Just take out the solar zenith correction
    # part. Previously, solar zenith correction was being applied twice, then we were
    # off by factor of pi/irrad. Now we should be good!
    # ref[rad > 0] = np.pi * rad[rad > 0] /
    #                                 (irrad * np.cos((np.pi / 180) * sun_zen[rad > 0]))
    ref[rad > 0] = np.pi * rad[rad > 0] / irrad
    ref[ref < 0] = 0
    ref[ref > 1] = 1
    ref[sun_zen > 95] = -999.0
    ref[sun_zen <= -999] = -999.0
    return ref


def radToBT(rad, platform, band):
    """Convert rad to BT."""
    c1 = 1.19104e-05
    c2 = 1.43877
    wn = IR_CALIB[platform][band]["wn"]
    a = IR_CALIB[platform][band]["a"]
    b = IR_CALIB[platform][band]["b"]
    temp = np.full_like(rad, -999.0)
    temp[rad > 0] = ((c2 * wn) / np.log(1 + wn**3 * c1 / rad[rad > 0]) - b) / a
    return temp


class Chan(object):
    """Channel class."""

    _good_names = [
        "B01Rad",
        "B01Ref",  # VIS0.6 Cloud mapping
        "B02Rad",
        "B02Ref",  # VIS0.8 Vegetation index
        "B03Rad",
        "B03Ref",  # NIR1.6 Cloud / snow discrimination
        "B04Rad",
        "B04BT",  # IR3.9  Atmospheric window
        "B05Rad",
        "B05BT",  # IR6.2  Water vapour channel (Upper atmosphere)
        "B06Rad",
        "B06BT",  # IR7.3  Water vapour channel (Lower atmosphere)
        "B07Rad",
        "B07BT",  # IR8.7  Atmospheric window
        "B08Rad",
        "B08BT",  # IR9.7  Ozone channel
        "B09Rad",
        "B09BT",  # IR10.8 Atmospheric window
        "B10Rad",
        "B10BT",  # IR12.0 Atmospheric window
        "B11Rad",
        "B11BT",
    ]  # IR13.4 Carbon dioxide channel

    def __init__(self, name):
        """Initialize Chan object."""
        if "B12" in name:
            raise HritError(
                "Channel 12 (High Resolution Visible) currently not handled."
            )
        if name not in self._good_names:
            raise ValueError("Unknown channel name: {}".format(name))
        self._name = name
        self._band = name[0:3]
        self._type = name[3:]

    @property
    def name(self):
        """Name property."""
        return self._name

    @property
    def band(self):
        """Band property."""
        return self._band

    @property
    def band_num(self):
        """Band number property."""
        return int(self._band[1:])

    @property
    def type(self):
        """Type property."""
        return self._type


class ChanList(object):
    """ChanList Class."""

    def __init__(self, chans):
        """Initialize ChanList object."""
        chans = set(chans)
        self._info = {"chans": [Chan(chan) for chan in chans]}
        self._info["names"] = list(set([chan.name for chan in self.chans]))
        self._info["bands"] = list(set([chan.band for chan in self.chans]))
        self._info["types"] = list(set([chan.type for chan in self.chans]))

    @property
    def chans(self):
        """Chans property."""
        return self._info["chans"]

    @property
    def names(self):
        """Names property."""
        return self._info["names"]

    @property
    def bands(self):
        """Bands property."""
        return self._info["bands"]

    @classmethod
    def _all_types_for_bands(cls, bands):
        """List all types for bands."""
        good_names = Chan._good_names
        chans = set()
        for chan in good_names:
            for band in bands:
                if band in chan:
                    chans.add(chan)
        return cls(chans)


def call_single_time(
    fnames, metadata_only=False, chans=None, area_def=None, self_register=False
):
    """Read SEVIRI hrit data products.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    gvars = {}
    datavars = {}
    adname = "undefined"
    # Remove any HRV files from file list
    # See note 1 at top of module

    fnames = [fname for fname in fnames if not any(val in fname for val in ["HRV"])]
    if not fnames:
        raise NoValidFilesError("No files found in list, skipping")

    # Check inputs
    if self_register and self_register != "LOW":
        raise XritError(
            "Unknown resolution supplied to self_register: {}".format(self_register)
        )
    if area_def:
        try:
            adname = area_def.name
        except AttributeError:
            TypeError("Keyword area_def must be of type Sector.")
    else:
        adname = "FULL_DISK"

    import xarray

    xarray_obj = xarray.Dataset()

    # Gather top-level metadata. MUst pass ALL fnames to make sure we
    # use a datafile, and not pro or epi (they do not contain projection
    # information)
    xarray_obj.attrs = get_top_level_metadata(fnames, area_def)

    # chans == [] specifies we don't want to read ANY data, just metadata.
    # chans == None specifies that we are not specifying a channel list,
    #               and thus want ALL channels.
    if metadata_only:
        # If NO CHANNELS were specifically requested, just return at this
        # point with the metadata fields populated. A dummy SciFile dataset
        # will be created with only metadata. This is for checking what
        # platform/source combination we are using, etc.
        return {"METADATA": xarray_obj}

    # Create file objects for each input file and organize
    dfs = {}
    pro = None
    epi = None
    sdt = None
    imgf = None
    all_segs = set()
    from struct import error as structerror

    for fname in fnames:
        try:
            df = HritFile(fname)
        except structerror:
            # Don't fail altogether if there is one bad file
            LOG.exception("FAILED reader %s, skipping", fname)
            continue
        # Ensure all files have same start datetime
        if not sdt:
            sdt = df.start_datetime
        if df.start_datetime != sdt:
            raise HritError(
                "Start date time does not match for all files: {}".format(fnames)
            )
        # Get prologue
        if df.file_type == "prologue":
            pro = df.prologue
            dt = xarray_obj.attrs["start_datetime"]
            # xarray_obj.attrs['prologue'] = pro
            for poly in df.prologue["satelliteStatus"]["orbit"]["orbitPolynomial"]:
                if dt <= poly["endTime"] and dt >= poly["startTime"]:
                    LOG.info("Calculating x/y/z satellite location")

                    st = poly["startTime"]
                    et = poly["endTime"]
                    xcoef = poly["X"]
                    ycoef = poly["Y"]
                    zcoef = poly["Z"]
                    x = calculate_chebyshev_polynomial(xcoef, st, et, dt)
                    y = calculate_chebyshev_polynomial(ycoef, st, et, dt)
                    z = calculate_chebyshev_polynomial(zcoef, st, et, dt)
                    xarray_obj.attrs["satECF_m"] = {}
                    xarray_obj.attrs["satECF_m"]["x"] = x * 1000
                    xarray_obj.attrs["satECF_m"]["y"] = y * 1000
                    xarray_obj.attrs["satECF_m"]["z"] = z * 1000

        # Get epilogue
        elif df.file_type == "epilogue":
            epi = df.epilogue
            xarray_obj.attrs["epilogue"] = epi
        # Store data files, organized by band, then segment
        elif df.file_type == "image":
            # Ensure all files have the same geolocation information
            if not imgf:
                imgf = df
            if not compare_dicts(
                imgf.geolocation_metadata,
                df.geolocation_metadata,
                skip=["line_offset", "sample_offset"],
            ):
                raise HritError(
                    "Geolocation metadata do not match for image files: {}".format(
                        fnames
                    )
                )
            # Initialize band info with None for eight segments
            if df.band not in dfs:
                dfs[df.band] = {num: None for num in range(1, 9)}
            dfs[df.band][df.segment] = df
            all_segs.add(df.segment)
        else:
            LOG.warning("Unhandled file type encountered: {}".format(df.file_type))
    if not pro:
        raise HritError("No prologue file found")

    # If specific channels were requested, check them against the input data
    if chans:
        chlist = ChanList(list(set(chans) - set(geolocation_variable_names)))
        for chan in chlist.chans:
            if chan.band not in dfs.keys():
                raise ValueError(
                    "Requested channel {} not found in input data.".format(chan.name)
                )
    # If no specific channels were requested, get everything
    else:
        chlist = ChanList._all_types_for_bands(dfs.keys())

    xarray_obj.attrs["datavars"] = {}
    # Gather geolocation data
    # Assume the datetime is the same for all resolution.  Not true, but close enough.
    # This saves us from having slightly different solar angles for each channel.
    gmd = _get_geolocation_metadata(pro, xarray_obj.attrs)
    fldk_lats, fldk_lons = get_latitude_longitude(gmd, BADVALS, area_def)
    gvars[adname] = get_geolocation(sdt, gmd, fldk_lats, fldk_lons, BADVALS, area_def)

    # Drop files for channels other than those requested and decompress
    outdir = os.path.join(
        gpaths["LOCALSCRATCH"],
        xarray_obj.attrs["source_name"],
        xarray_obj.attrs["platform_name"],
        xarray_obj.attrs["start_datetime"].strftime("%Y%m%d%H%M"),
    )

    from geoips.filenames.base_paths import make_dirs

    make_dirs(outdir)

    # Can not change dictionary size during iteration for Python 3
    skip_bands = []
    for band in dfs.keys():
        if band not in chlist.bands:
            skip_bands += [band]
            # dfs.pop(band)
            continue
        else:
            # dfs[band] = {seg: df.decompress(outdir) for seg, df in dfs[band].items()}
            skip_segs = []
            for seg, df in dfs[band].items():
                if df is None:
                    LOG.error(
                        "FAILED READ ABOVE, SKIPPING TRYING TO DECOMPRESS FILE %s %s",
                        band,
                        seg,
                    )
                    skip_segs += [seg]
                    # dfs[band].pop(seg)
                    continue
                try:
                    dfs[band][seg] = df.decompress(outdir)
                except HritError as resp:
                    skip_segs += [seg]
                    LOG.error(
                        "FAILED DECOMPRESSING, %s, SKIPPING FILE %s", str(resp), df.name
                    )
            for skip_seg in skip_segs:
                dfs[band].pop(skip_seg)
    for skip_band in skip_bands:
        dfs.pop(skip_band)

    # Create data arrays for requested data and read count data
    num_lines = pro["imageDescription"]["referenceGridVIS_IR"]["numberOfLines"]
    num_samples = pro["imageDescription"]["referenceGridVIS_IR"]["numberOfColumns"]
    count_data = {}
    annotation_metadata = {}
    for band in chlist.bands:
        if len(dfs[band].items()) == 0:
            LOG.error("No data in band %s, SKIPPING", band)
            continue
        # Create empty full-disk array for this channel
        data = np.full((num_lines, num_samples), -999.9, dtype=float)
        # Read data into data array
        for seg, df in dfs[band].items():
            seg_num_lines = df.metadata["block_1"]["num_lines"]
            start_line = seg_num_lines * (seg - 1)
            end_line = seg_num_lines * seg
            try:
                data[start_line:end_line, 0:] = df._read_image_data()
            except ValueError as resp:
                LOG.error("FAILED READING SEGMENT, SKIPPING %s" % (resp))
        LOG.info("Read band %s %s" % (band, df.annotation_metadata["band"]))
        if "Lines" in gvars[adname]:
            count_data[band] = data[gvars[adname]["Lines"], gvars[adname]["Samples"]]
        else:
            count_data[band] = data
        annotation_metadata[band] = df.annotation_metadata

    datavars[adname] = {}
    radiances = {}
    image_cal = pro["radiometricProcessing"]["level15ImageCalibration"]
    for chan in chlist.chans:
        counts = count_data[chan.band]
        if chan.band not in radiances:
            band_cal = image_cal[chan.band_num - 1]
            offset = band_cal["offset"]
            slope = band_cal["slope"]
            radiances[chan.band] = countsToRad(counts, slope, offset)
        if chan.type == "Rad":
            datavars[adname][chan.name] = radiances[chan.band]
        if chan.type == "Ref":
            LOG.info(
                "Calculating reflectances for %s, data range %f to %f",
                chan.band,
                radiances[chan.band].min(),
                radiances[chan.band].max(),
            )
            datavars[adname][chan.name] = radToRef(
                radiances[chan.band],
                gvars[adname]["solar_zenith_angle"],
                xarray_obj.attrs["satellite_name"],
                chan.band,
            )
            LOG.info(
                "Final reflectances for %s, data range %f to %f",
                chan.band,
                datavars[adname][chan.name].min(),
                datavars[adname][chan.name].max(),
            )
        if chan.type == "BT":
            LOG.info(
                "Calculating brightness temperatures for %s, data range %f to %f",
                chan.band,
                radiances[chan.band].min(),
                radiances[chan.band].max(),
            )
            datavars[adname][chan.name] = radToBT(
                radiances[chan.band], xarray_obj.attrs["satellite_name"], chan.band
            )
            LOG.info(
                "Final brightness temperatures for %s, data range %f to %f",
                chan.band,
                datavars[adname][chan.name].min(),
                datavars[adname][chan.name].max(),
            )
        if adname not in xarray_obj.attrs["datavars"].keys():
            xarray_obj.attrs["datavars"][adname] = {}
        if chan.name not in xarray_obj.attrs["datavars"].keys():
            xarray_obj.attrs["datavars"][adname][chan.name] = {}

        xarray_obj.attrs["datavars"][adname][chan.name]["wavelength"] = float(
            annotation_metadata[chan.band]["band"][3:5]
            + "."
            + annotation_metadata[chan.band]["band"][5:]
        )

    gvars[adname]["latitude"] = np.ma.masked_less_equal(gvars[adname]["latitude"], -999)
    toplat = gvars[adname]["latitude"][np.ma.where(gvars[adname]["latitude"])][0]
    bottomlat = gvars[adname]["latitude"][np.ma.where(gvars[adname]["latitude"])][-1]

    for var in gvars[adname].keys():
        if toplat < bottomlat:
            gvars[adname][var] = np.ma.masked_less_equal(
                np.flipud(gvars[adname][var]), -999
            )
        else:
            gvars[adname][var] = np.ma.masked_less_equal(gvars[adname][var], -999)

        if "satellite_zenith_angle" in gvars[adname].keys():
            gvars[adname][var] = np.ma.masked_where(
                gvars[adname]["satellite_zenith_angle"] > 75, gvars[adname][var]
            )

    for var in datavars[adname].keys():
        if toplat < bottomlat:
            datavars[adname][var] = np.ma.masked_less_equal(
                np.flipud(datavars[adname][var]), -999
            )
        else:
            datavars[adname][var] = np.ma.masked_less_equal(datavars[adname][var], -999)

        if "satellite_zenith_angle" in gvars[adname].keys():
            datavars[adname][var] = np.ma.masked_where(
                gvars[adname]["satellite_zenith_angle"] > 75, datavars[adname][var]
            )

    xarray_objs = {}
    for dsname in datavars.keys():
        xobj = xarray.Dataset()
        xobj.attrs = xarray_obj.attrs.copy()
        for varname in datavars[dsname].keys():
            xobj[varname] = xarray.DataArray(datavars[dsname][varname])
        for varname in gvars[dsname].keys():
            xobj[varname] = xarray.DataArray(gvars[dsname][varname])
        if hasattr(xobj.attrs, "area_definition") and xobj.area_definition is not None:
            xobj.attrs["interpolation_radius_of_influence"] = max(
                xobj.area_definition.pixel_size_x, xobj.area_definition.pixel_size_y
            )
        else:
            xobj.attrs["interpolation_radius_of_influence"] = 10000
        xarray_objs[dsname] = xobj

    xarray_objs["METADATA"] = list(xarray_objs.values())[0][[]]
    LOG.info("Done reading SEVIRI data for {}".format(adname))
    LOG.info("")

    return xarray_objs


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read SEVIRI hrit data products.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    # Remove any HRV files from file list. Need to do this here instead of in
    # 'call_single_time' as that will initially be fed one file at a time. If that file
    # is a HRV file, then this will result in an empty list and an error thrown in
    # 'get_top_level_metadata'.
    # See note 1 at top of module
    fnames = [fname for fname in fnames if not any(val in fname for val in ["HRV"])]

    return readers.read_data_to_xarray_dict(
        fnames,
        call_single_time,
        metadata_only,
        chans,
        area_def,
        self_register,
    )



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Standard GeoIPS static title production."""

# Python Standard Libraries
import logging

LOG = logging.getLogger(__name__)

interface = "title_formatters"
family = "standard"
name = "static_standard"


def call(
    area_def,
    xarray_obj,
    product_name_title,
    product_datatype_title=None,
    bg_xarray=None,
    bg_product_name_title=None,
    bg_datatype_title=None,
    title_copyright=None,
):
    """Generate standard GeoIPS formatted title."""
    title_line1 = "{0} {1}".format(product_datatype_title, product_name_title)
    title_line2 = "{0}".format(xarray_obj.start_datetime.strftime("%Y/%m/%d %H:%M:%SZ"))
    if bg_xarray is not None:
        title_line3 = "{0} {1} at {2}".format(
            bg_datatype_title,
            bg_product_name_title,
            bg_xarray.start_datetime.strftime("%Y-%m-%d %H:%M:%S"),
        )
        # title_string = f'{title_line1}\n{title_line2}\n{title_line3}\n
        #                                                             {title_copyright}'
        title_string = f"{title_line1}\n{title_line2} {title_copyright}\n{title_line3}"
    else:
        # title_string = f'{title_line1}\n{title_line2}\n{title_copyright}'
        title_string = f"{title_line1}\n{title_line2} {title_copyright}"
    LOG.info("Not dynamic, using standard title_string: %s", title_string)

    return title_string



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Products interface module."""

import logging
from geoips.geoips_utils import merge_nested_dicts
from geoips.interfaces.base import YamlPluginValidator, BaseYamlInterface
from geoips.interfaces.yaml_based.product_defaults import product_defaults
from jsonschema.exceptions import ValidationError

LOG = logging.getLogger(__name__)


class ProductsPluginValidator(YamlPluginValidator):
    """Validator for Products plugins.

    This differs from other validators for two reasons:

    1. Most plugins are identified solely by 'name'. Products are identified by
       'source_name' and 'name.
    2. Most plugins supply their 'family' directly. Products can supply it directly, but
       can, alternatively, specify a 'product_defaults' plugin from which to pull
       'family' and most other properties. This validator handles filling in a product
       plugin based on its specified product defaults plugin.
    """

    def validate(self, plugin, validator_id=None):
        """Validate a Products plugin against the relevant schema.

        The relevant schema is determined based on the interface and family of the
        plugin.
        """
        if "product_defaults" in plugin:
            if "family" in plugin:
                raise ValidationError(
                    "Properties 'family' and 'product_defaults' are mutually exclusive."
                )
            self.validate_product(plugin)
        else:
            try:
                plugin = super().validate(plugin, validator_id=validator_id)
            except ValidationError as resp:
                raise ValidationError(
                    f"{resp}: Trouble validating plugin '{plugin.get('name')}'\n"
                    f"in interface '{plugin.get('interface')}'\n"
                    f"located at '{plugin.get('abspath')}'\n"
                    f"from package '{plugin.get('package')}' "
                ) from resp

        return plugin

    def validate_product(self, product):
        """Validate single product."""
        LOG.debug("In validate product")
        if "family" in product:
            LOG.debug("Validating family-based product")
            family = product["family"]
            try:
                spec_validator = self.validators[f"product_defaults.specs.{family}"]
            except KeyError:
                raise ValidationError(
                    f"No product_defaults spec for family '{family}'", instance=product
                )
            spec_validator.validate(product["spec"])
        elif "product_defaults" in product:
            defaults = product_defaults.get_plugin(product.pop("product_defaults"))
            product["family"] = defaults["family"]

            LOG.debug("Validating product_defaults-based product")
            # This updates missing values in spec from defaults but leaves existing
            # values alone. Using update here ensures that we're updating in-place
            # rather than creating a new dictionary.
            merge_nested_dicts(product, defaults)
        else:
            raise ValidationError(
                f"Product '{product['name']}' did not specify either "
                f"'family' or 'product_defaults'."
            )
        return product


class ProductsInterface(BaseYamlInterface):
    """GeoIPS interface for Products plugins."""

    name = "products"
    validator = ProductsPluginValidator()

    def _create_registered_plugin_names(self, yaml_plugin):
        """Create a plugin name for plugin registry.

        This name is a tuple containing source_name and name.
        Overrides the same method from YamlPluginValidator.
        """
        names = []
        for source_name in yaml_plugin["source_names"]:
            names += [(source_name, yaml_plugin["name"])]
        return names

    def get_plugin_metadata(self, source_name, name):
        """Retrieve a product plugin's metadata.

        Where the metadata of the plugin matches the plugin's corresponding entry in the
        plugin registry.

        Parameters
        ----------
        source_name: str
            - The source (sensor) which this product is derived from.
        name: str
            - The name of the product plugin whose metadata we'd like to retrieve.

        Returns
        -------
        metadata: dict
            - A dictionary of metadata for the requested plugin.
        """
        return super().get_plugin_metadata((source_name, name))

    def get_plugin(
        self, source_name, name, product_spec_override=None, rebuild_registries=None
    ):
        """Retrieve a Product plugin by source_name, name, and product_spec_override.

        If product_spec_override dict is passed, values contained within
        product_spec_override will be used in place of those found in products
        list and product_defaults.

        product_spec_override[product_name] matches the format of the product
        "spec" field.

        Additionally, if the special key product_spec_override["all"] is included,
        it will apply to all products not specified by name within the dictionary.

        Parameters
        ----------
        source_name : str
            - The name the source which the product is derived from.
        name : str
            - The name the desired plugin.
        dict : str
            - Dictionary specifying what information of the product's spec that is to be
              overridden at runtime.
        rebuild_registries: bool (default=None)
            - Whether or not to rebuild the registries if get_plugin fails. If set to
              None, default to what we have set in geoips.filenames.base_paths, which
              defaults to True. If specified, use the input value of rebuild_registries,
              which should be a boolean value. If rebuild registries is true and
              get_plugin fails, rebuild the plugin registry, call then call
              get_plugin once more with rebuild_registries toggled off, so it only gets
              rebuilt once.
        """
        prod_plugin = super().get_plugin((source_name, name), rebuild_registries)
        if product_spec_override is not None:
            # Default to no override arguments
            override_args = {}
            # If available, use the current product's override values
            if name in product_spec_override:
                override_args = product_spec_override[name]
            # Otherwise, if "all" specified, use those override values
            elif "all" in product_spec_override:
                override_args = product_spec_override["all"]
            replace_arg = False
            # NOTE: this is a top-level field, where if you set
            # product_spec_override:
            #   spec:
            #     replace: true
            # It will automatically replace ALL fields found in
            # the original product spec and also found in the
            # override with what is specified in the override
            # in its entirety, without merging.  This is not
            # terribly useful overall - we probably want this
            # sort of capability in the end, but more flexible
            # and able to be applied to only specific fields,
            # etc.  This is a brute force method to at least
            # allow overriding entire fields.
            if product_spec_override.get("replace"):
                replace_arg = True
            prod_plugin["spec"] = merge_nested_dicts(
                prod_plugin["spec"], override_args, in_place=False, replace=replace_arg
            )

        return prod_plugin

    def get_plugins(self):
        """Retrieve a plugin by name."""
        plugins = []
        for source_name in self.plugin_registry.registered_plugins["yaml_based"][
            self.name
        ].keys():
            for subplg_name in self.plugin_registry.registered_plugins["yaml_based"][
                self.name
            ][source_name].keys():
                plugins.append(self.get_plugin(source_name, subplg_name))
        return plugins

    def plugin_is_valid(self, source_name, name):
        """Test that the named plugin is valid."""
        try:
            self.get_plugin(source_name, name)
            return True
        except ValidationError:
            return False

    def test_interface(self):
        """Test interface method."""
        plugins = self.get_plugins()
        all_valid = self.plugins_all_valid()
        family_list = []
        plugin_ids = {}
        for plugin in plugins:
            if plugin.family not in family_list:
                family_list.append(plugin.family)
                plugin_ids[plugin.family] = []
            plugin_ids[plugin.family].append(plugin.id)

        output = {
            "all_valid": all_valid,
            "by_family": plugin_ids,
            "validity_check": {},
            "family": {},
            "func": {},
            "docstring": {},
        }
        for curr_family in plugin_ids:
            for curr_id in plugin_ids[curr_family]:
                output["validity_check"][curr_id] = self.plugin_is_valid(*curr_id)
                output["func"][curr_id] = self.get_plugin(*curr_id)
                output["family"][curr_id] = curr_family
                output["docstring"][curr_id] = output["func"][curr_id].docstring
        return output


products = ProductsInterface()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for the interface method 'get_plugin_metadata'."""

import pytest

from geoips import interfaces
from geoips.errors import PluginRegistryError


def generate_id(iname_pname):
    """Generate an ID for the test-arguments provided."""
    interface = iname_pname[0]
    plugin_name = iname_pname[1]
    if isinstance(plugin_name, tuple):
        return f"{interface} " + " ".join(plugin_name)
    return f"{interface} {str(plugin_name)}"


int_plg_pairs = []

idict = interfaces.list_available_interfaces()

for itype in idict:
    for iname in idict[itype]:
        interface = getattr(interfaces, iname)
        iregistry = interface.plugin_registry.registered_plugins.get(itype, {}).get(
            iname
        )
        if iregistry:
            if iname != "products":
                for pname in iregistry:
                    int_plg_pairs.append([iname, pname])
            else:
                for sname in iregistry:
                    for pname in iregistry[sname]:
                        int_plg_pairs.append([iname, (sname, pname)])

int_plg_pairs.append(["fake_interface", "fake_plugin"])
int_plg_pairs.append(["algorithms", 2])
int_plg_pairs.append(["algorithms", "non_existent_plugin"])


@pytest.mark.parametrize(
    "iname_pname",
    int_plg_pairs,
    ids=generate_id,
)
def test_get_plugin_metadata(iname_pname):
    """Test that <interface_name>.get_plugin_metadata(<name>) works as expected.

    Includes failing cases as well. This should hit 100% coverage for the new code.
    """
    iname = iname_pname[0]
    pname = iname_pname[1]

    if iname == "fake_interface":
        interface = getattr(interfaces, "algorithms")
        interface.interface_type = "some_fake_type"
        # Failing case 1
        with pytest.raises(KeyError):
            md = interface.get_plugin_metadata(pname)
        # Make sure to reset the interface type; algorithms will need this later on
        interface.interface_type = "module_based"
        return
    else:
        interface = getattr(interfaces, iname)

    if isinstance(pname, tuple):
        md = interface.get_plugin_metadata(*pname)
    else:
        # Failing case 2
        if pname == 2:
            with pytest.raises(TypeError):
                md = interface.get_plugin_metadata(pname)
            return
        # Failing case 3
        elif pname == "non_existent_plugin":
            with pytest.raises(PluginRegistryError):
                md = interface.get_plugin_metadata(pname)
            return
        # Any other case should pass
        else:
            md = interface.get_plugin_metadata(pname)

    # All of these keys should exist for every plugin type
    for key in [
        "docstring",
        "family",
        "interface",
        "package",
        "plugin_type",
        "relpath",
    ]:
        assert key in md
